﻿//
// Myriadbits Disk space analyzer 
// Find free space on your NTFS disk
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
//using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;
using Myriadbits.NTFS;


namespace Myriadbits.DiskSpaceAnalyzer
{
	/// <summary>
	/// The main form of the DiskSpaceAnalyzer
	/// </summary>
	public partial class MainForm : Form
	{
		// Static colors
		public static Color m_textColorZeroSize = Color.FromArgb(160, 160, 160);
		public static Color m_textColor = Color.FromArgb(50, 50, 50);

		private const int s_BarWidthPercentage = 33; // The percentage bar is x% of the width of the control

		private bool m_isClosing = false;
		private ViewOptions m_currentViewOptions = ViewOptions.Percentage;
		private ViewOptionsEx m_currentViewOptionsEx = ViewOptionsEx.Size;
		
		private Folder m_currentSelectedFolder = null;
		private Myriadbits.NTFS.File m_currentSelectedFile = null;
		private string m_currentSelectedPath = "";

		private Color m_textColorFileInfo = Color.FromArgb(109, 109, 109);
		
		private Color m_barColorMain = Color.FromArgb(100, 0, 190, 50);
		private Color m_barColorSecondary = Color.FromArgb(100, 0, 250, 50);
		private Color m_barColorRoot = Color.FromArgb(100, 0, 50, 190);

		private Color m_barBorderColor = Color.FromArgb(60, 50, 50, 50);
		private Color m_barBorderColorVague = Color.FromArgb(20, 50, 50, 50);

		private Font m_treeFont = new System.Drawing.Font("Microsoft Sans Serif", 8);
		private Font m_treeFontBold = new System.Drawing.Font("Microsoft Sans Serif", 8, FontStyle.Bold);
		private Font m_treeFontSmall = new System.Drawing.Font("Microsoft Sans Serif", 7);

		private Dictionary<TreeViewMode, string> m_viewingModes = new Dictionary<TreeViewMode, string>();

		private UnitTypes m_currentUnit = 0;

		private ImageList m_imageFoldersNormal = new ImageList();
		private ImageList m_imageFoldersOpen = new ImageList();

		private List<long> m_vecFreeSpacesPerDrive = new List<long>();
		private bool m_fDeletingFile = false;
		private bool m_fShowingRefreshQuestion = false;

		private ulong m_maxLogValue = TreeBuilder.GIGABYTE100;

		private StringFormat m_treeTextFormat = new StringFormat(StringFormatFlags.NoWrap);

		/// <summary>
		/// Constructor
		/// </summary>
		public MainForm()
		{
			InitializeComponent();
		}


		/// <summary>
		/// Load this form
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void MainForm_Load(object sender, EventArgs e)
		{
			// Refresh settings
			LoadUserSettings();

			// Bind combobox to dictionary
			m_viewingModes.Add(TreeViewMode.AllFiles, "Show all files and folders");
			m_viewingModes.Add(TreeViewMode.HideEmptyFolders, "Show all folders except empty folders");
			//m_viewingModes.Add(TreeViewMode.FileLargest10, "Show largest 10 files (autoexpand)");
			//m_viewingModes.Add(TreeViewMode.FileLargest100, "Show largest 100 files (autoexpand)");
			//m_viewingModes.Add(TreeViewMode.FileLargest1000, "Show largest 1000 files");
			m_viewingModes.Add(TreeViewMode.FolderLargerThen1MB, "Show all folders larger then 1 MB");
			m_viewingModes.Add(TreeViewMode.FolderLargerThen10MB, "Show all folders larger then 10 MB");
			m_viewingModes.Add(TreeViewMode.FolderLargerThen100MB, "Show all folders larger then 100 MB");
			m_viewingModes.Add(TreeViewMode.FolderLargerThen1GB, "Show all folders larger then 1 GB");
			m_viewingModes.Add(TreeViewMode.FolderLargerThen10GB, "Show all folders larger then 10 GB");
			m_viewingModes.Add(TreeViewMode.FolderLargerThen100GB, "Show all folders larger then 100 GB");
			m_viewingModes.Add(TreeViewMode.FileLargerThen1MB, "Only show files larger then 1 MB");
			m_viewingModes.Add(TreeViewMode.FileLargerThen10MB, "Only show files larger then 10 MB");
			m_viewingModes.Add(TreeViewMode.FileLargerThen100MB, "Only show files larger then 100 MB (autoexpand)");
			m_viewingModes.Add(TreeViewMode.FileLargerThen1GB, "Only show files larger then 1 GB (autoexpand)");
			m_viewingModes.Add(TreeViewMode.FileLargerThen10GB, "Only show files larger then 10 GB (autoexpand)");
			m_viewingModes.Add(TreeViewMode.FileLargerThen100GB, "Only show files larger then 100 GB (autoexpand)");

			// Bind the list to the combobox
			cmbOptions.Enabled = false;
			cmbOptions.DataSource = new BindingSource(m_viewingModes, null);
			cmbOptions.DisplayMember = "Value";
			cmbOptions.ValueMember = "Key";
			cmbOptions.Enabled = true;

			// Select the last selected item
			int newViewModeIndex = DiskSpaceAnalyzer.Properties.Settings.Default.Main_ViewMode;
			if (newViewModeIndex < 0 || newViewModeIndex >= m_viewingModes.Count())
				newViewModeIndex = 0;
			cmbOptions.SelectedIndex = newViewModeIndex;

			// Set all imagelists to the correct bitdepth
			m_imageFoldersNormal.ColorDepth = ColorDepth.Depth32Bit;
			m_imageFoldersOpen.ColorDepth = ColorDepth.Depth32Bit;

			// Link the top list
			this.lvwTopFiles.Initialize(DiskSpaceAnalyzer.Properties.Settings.Default.Main_TopListNumber);

			// Start the tree
			InitializeTree();
			

			// Store all free spaces for all drives
			m_vecFreeSpacesPerDrive.Clear();
			DriveInfo[] allDrives = DriveInfo.GetDrives();
			foreach (DriveInfo d in allDrives)
				m_vecFreeSpacesPerDrive.Add(d.AvailableFreeSpace);

			// Check if the free space has changed every second
			this.timerCheckFree.Start();

			// Set the initial viewing modes
			ViewOptions lastViewOptions;
			ViewOptionsEx lastViewOptionsEx;
			if (!Enum.TryParse<ViewOptions>(DiskSpaceAnalyzer.Properties.Settings.Default.Main_ViewOptions, out lastViewOptions))
				lastViewOptions = ViewOptions.Percentage;
			if (!Enum.TryParse<ViewOptionsEx>(DiskSpaceAnalyzer.Properties.Settings.Default.Main_ViewOptionsEx, out lastViewOptionsEx))
				lastViewOptionsEx = ViewOptionsEx.Size;

			ChangeViewingMode(lastViewOptions);
			ChangeViewingModeEx(lastViewOptionsEx);

			TreeViewMode newViewMode = ((KeyValuePair<TreeViewMode, string>)cmbOptions.SelectedItem).Key;
			this.lvwTreeFiles.ChangeViewMode(newViewMode);
			this.lvwTopFiles.ChangeViewMode(newViewMode);
			this.lvwSearchResults.ChangeViewMode(newViewMode);

			this.lvwTreeFiles.LinkStatusInfo(this.sslDriveInfo);

			// Set the tree text formating options
			this.m_treeTextFormat.Trimming = StringTrimming.EllipsisCharacter;
		}


		/// <summary>
		/// After showing the main screen, show the startup screen
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void MainForm_Shown(object sender, EventArgs e)
		{
			ShowScanDialog();
		}


		/// <summary>
		/// Show the scanning dialog
		/// </summary>
		private void ShowScanDialog(bool autoStart = false)
		{
			this.Enabled = false;
			StartForm sf = new StartForm(autoStart);
			if (sf.ShowDialog() == System.Windows.Forms.DialogResult.OK)
			{
				this.treeMain.Nodes.Clear();
				this.lvwTopFiles.ClearList();
				
				TreeViewMode viewMode = ((KeyValuePair<TreeViewMode, string>)cmbOptions.SelectedItem).Key;
				foreach (DriveData dd in sf.DriveData)
				{
					// Build the tree					
					TreeNode tnroot = new TreeNode(dd.DriveDisplayName);
					RootFolder rootFolder = new RootFolder(dd.Parser.RootFolder, dd.DriveDisplayName);
					rootFolder.Builder = new TreeBuilder(tnroot, rootFolder, viewMode, this.m_currentSelectedFolder);
					this.treeMain.Nodes.Add(tnroot);

					// Add the rootfolder to the toplist
					this.lvwTopFiles.AddRoot(rootFolder);
				}


				if (this.m_currentSelectedFolder == null)
				{
					// No current selection, select the root
					if (this.treeMain.Nodes.Count > 0)
					{
						this.treeMain.SelectedNode = this.treeMain.Nodes[0];
					}
				}

				// Refresh the toplist
				this.lvwTopFiles.RefreshFileList();
			}
			this.Enabled = true;
		}


		/// <summary>
		/// Fill the tree with all (NTFS) harddisk
		/// </summary>
		private void InitializeTree()
		{
			this.treeMain.Nodes.Clear();

			DriveInfo[] allDrives = DriveInfo.GetDrives();
			foreach (DriveInfo d in allDrives)
			{
				if ((ulong) d.TotalSize > this.m_maxLogValue)
				{
					this.m_maxLogValue = (ulong) d.TotalSize;
				}

				string driveName = d.Name.Substring(0, 1);
				string volname = string.IsNullOrEmpty(d.VolumeLabel) ? "Local Disk" : d.VolumeLabel;
				string driveSt = string.Format("{0} ({1})", volname, d.Name);
				if (NTFSReader.IsDriveNTFS(driveName))
				{
					RootFolder folder = new RootFolder();
					folder.Size = (ulong)d.TotalSize;
					folder.Name = driveName;
					folder.Builder = null;

					TreeNode tn = new TreeNode(driveSt);					
					tn.Tag = folder;
					this.treeMain.Nodes.Add(tn);
				}
				else
				{
					TreeNode tn = new TreeNode(driveName + " (No NTFS!)");
					tn.Tag = null;
					this.treeMain.Nodes.Add(tn);
				}
			}
		}


		// TODO		
		// - Add file icons in search and top lists
		// - Minor: Add keyboard shortcuts to menus		
		// - Minor: After Refresh dialog => use the same current folder
		// - PUBLISH VERSION!!!!!!
		//
		// 
		// Features:
		// - Implement FAT32
		// - Show extra NTFS information regarding the selected file (lower left?) Only in advanced mode?
		// - Show bar in file view
		// - Show tree root lines? => Might be too busy

		// DONE: - Make Selection Ok for different disks!! => depends on show-all-folder stuff
		// DONE: - Delete file -> Also remove file from top-largest-list
		// DONE: - Add sorting to columns in largest X files tab
		// DONE: - Seperate tab page with largest X files (listview with path + file + size + menu with delete)
		// DONE: - Add sort options to the file list (going to get tricky since the list is virtual....)
		// DONE: - Remember current view mode and options
		// DONE: - Right mouse button menus inside normal menu
		// DONE: - In file view folders hidden die niet gezien mogen worden door instelling (hide empty folder)
		// DONE: - Huidige vrije ruimte altijd in scherm
		// DONE: - Auto Refresh vragen bij vrije ruimte verandering
		// REJECT - Bar afhankelijk van data vullen (onder 1% ander balken laten zien)
		// DONE: - Icons maken in toolbar
		// DONE: - Bar helemaal vooraan i.p.v. na map icon
		// DONE: - fileList disablen by geen files
		// DONE: - kolommen met info toevoegen by files
		// DONE: - Bar met kleur verloop (vooral by log bar) => log bar met tick marks is duidelijker
		// DONE: - Right mouse menu met:
		// DONE: - Windows properties
		// DONE: - Open explorer here
		// DONE: - Open command window here
		// DONE: - Expand all below + expand
		// DONE: - Delete file (in file list)
		// DONE: - Tooltip op filename in listview

		// REJECT: File names direct opslaan in Folder??  => YES, does NOT consume many bytes and is really fast!
		//	- Nadeel: veel meer geheugen -> Voordeel: Sorteren is mogelijk!
		//	- Voordeel: snel

		// DONE: File data opslaan voor huidige folder
		//  - Redelijk snel, veel minder geheugen

		// DONE:
		// * Startup scherm verder afmaken, met vraag welke drives gescanned moeten worden


		/// <summary>
		/// Draw-me-own progress bar + texts
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void treeMain_DrawNode(object sender, DrawTreeNodeEventArgs e)
		{
			if (!this.m_isClosing)
			{
				e.Graphics.FillRectangle(new SolidBrush(Color.White), e.Bounds);
				e.Graphics.SetClip(e.Bounds);

				if (e.Bounds.Height == 0)
					return;

				if (e.Node.IsSelected)
					e.Graphics.FillRectangle(new SolidBrush(SystemColors.GradientActiveCaption), e.Bounds);
			
				Rectangle bounds = e.Bounds;
				int totalBarWidth = s_BarWidthPercentage * e.Bounds.Width / 100;							
				int startOffsetBar = 0;
				int startOffsetExp = totalBarWidth + (e.Node.Level * treeMain.Indent);

				// Draw the expand button/triangle
				int imageIndex;
				if (e.Node.Nodes.Count > 0)
				{
					imageIndex = (e.Node.IsExpanded) ? 4 : 5;
					this.ilFolders.Draw(e.Graphics, new Point(startOffsetExp, e.Node.Bounds.Top), imageIndex);
				}

				imageIndex = (e.Node.Parent == null) ? 0 : 1;
				Font nameFont = m_treeFont;
				int vertTextOffsetSmall = 3;
				int vertTextOffsetNormal = 2;
				Folder folder = (Folder)e.Node.Tag;
				string iconType = "";
				SizeF textSize = new SizeF(0, 0);
				if (folder != null)
				{
					string text;					
					SolidBrush sb = null;
					double percentage1 = 0;
					double percentage2 = 0;

					switch (m_currentViewOptions)
					{
						//
						// TODO Make formatter classes that draw the data!
						//
						case ViewOptions.Percentage:
						case ViewOptions.PercentageLog:
						case ViewOptions.PercentageRelativeToParent:

							if (m_currentViewOptions == ViewOptions.PercentageLog)
							{
								percentage1 = folder.TotalSizePercentageLog(this.m_maxLogValue);
								percentage2 = folder.TotalOwnSizePercentageLog(this.m_maxLogValue);

							}
							else if (m_currentViewOptions == ViewOptions.PercentageRelativeToParent)
							{
								percentage1 = folder.ParentSizePercentage;
								percentage2 = folder.ParentOwnSizePercentage;
							}
							else
							{
								percentage1 = folder.TotalSizePercentage;
								percentage2 = folder.TotalOwnSizePercentage;
							}

							if (percentage1 < 0) percentage1 = 0;
							if (percentage2 < 0) percentage2 = 0;

							Color colMain = this.m_barColorMain;
							if (folder.ParentFolder == null)
							{
								// Root folder, always draw the used percentage
								DriveInfo info = new DriveInfo(folder.Name);
								if (info != null)
								{
									if (info.TotalSize != 0)
										percentage1 = ((double) folder.Size) / ((double) info.TotalSize);
									percentage2 = 0;
									colMain = this.m_barColorRoot;
								}								
								nameFont = m_treeFontBold; // Use bold for root folders
							}

							int textWidth = 36;
							totalBarWidth -= textWidth;

							// Draw the percentage bar
							drawBar(e, startOffsetBar + textWidth, e.Node.Bounds.Top + 4, totalBarWidth, 8, percentage1, percentage2, colMain, this.m_barColorSecondary);

							// Draw the tick marks on the bar
							if (m_currentViewOptions == ViewOptions.PercentageLog)
								drawLogMarks(e, startOffsetBar + textWidth, e.Node.Bounds.Top + 4, totalBarWidth, 8, this.m_maxLogValue);

							// Draw the percentage text
							if (percentage1 == 0)
							{
								sb = new SolidBrush(m_textColorZeroSize);
								text = string.Format("0%");
							}
							else
							{
								sb = new SolidBrush(m_textColor);
								text = string.Format("{0:0.0}%", percentage1 * 100.0);
							}
							e.Graphics.DrawString(text, this.m_treeFontSmall, sb, startOffsetBar, e.Node.Bounds.Top + vertTextOffsetSmall);

							startOffsetBar += totalBarWidth + textWidth;
							break;
					}

			

					//
					// Draw the text on the right side
					switch (m_currentViewOptionsEx)
					{
						case ViewOptionsEx.Size:
							text = this.m_currentUnit.GetUnitText(folder.Size);
							textSize = e.Graphics.MeasureString(text, this.m_treeFontSmall);
							if (folder.Size == 0)
								sb = new SolidBrush(m_textColorZeroSize);
							else
								sb = new SolidBrush(m_textColor);
							e.Graphics.DrawString(text, this.m_treeFontSmall, sb, e.Bounds.Right - textSize.Width - 8, e.Bounds.Top + vertTextOffsetSmall);
							break;

						case ViewOptionsEx.NumberOfFiles:
							// Draw the number of files
							if (folder.NumberOfFiles == 0)
							{
								sb = new SolidBrush(m_textColorZeroSize);
								text = string.Format("No files");
							}
							else
							{
								sb = new SolidBrush(m_textColor);
								text = string.Format("{0:### ### ### ### ###} Files", folder.NumberOfFiles);
							}
							textSize = e.Graphics.MeasureString(text, this.m_treeFontSmall);
							e.Graphics.DrawString(text, this.m_treeFontSmall, sb, e.Bounds.Right - textSize.Width - 8, e.Bounds.Top + vertTextOffsetSmall);
							break;

						case ViewOptionsEx.FileID:
							// Draw the number of files
							text = string.Format("{0:### ### ### ### ###}", folder.ReferenceNumber);
							textSize = e.Graphics.MeasureString(text, this.m_treeFontSmall);
							e.Graphics.DrawString(text, this.m_treeFontSmall, sb, e.Bounds.Right - textSize.Width - 8, e.Bounds.Top + vertTextOffsetSmall);
							break;
					}
				}


				int startOffsetText = startOffsetExp + 16;


				// Get the icon type of this folder
				iconType = FileIcon.GetIconType(folder.FullPath);

				// Add the icon to the imagelist (If not already present)		
				if (!string.IsNullOrEmpty(iconType))
				{
					if (!m_imageFoldersNormal.Images.ContainsKey(iconType))
					{
						// Load the icon type
						Icon iconNew = FileIcon.GetSmallIcon(folder.FullPath);
						if (iconNew != null)
							m_imageFoldersNormal.Images.Add(iconType, iconNew);
					}
				}


				// Now draw after the bar
				// Draw the folder icon (or file icon)
				startOffsetText += 4;
				if (!string.IsNullOrEmpty(iconType))
					m_imageFoldersNormal.Draw(e.Graphics, new Point(startOffsetText, e.Node.Bounds.Top + vertTextOffsetNormal), m_imageFoldersNormal.Images.IndexOfKey(iconType));
				
				// Draw the normal folder text
				startOffsetText += 16 + 2;
				Rectangle rectLayout = new Rectangle(startOffsetText, e.Node.Bounds.Top + vertTextOffsetNormal, e.Bounds.Width - startOffsetText - (int)textSize.Width - 8, e.Node.Bounds.Height);

				e.Graphics.DrawString(e.Node.Text, nameFont, new SolidBrush(Color.FromArgb(0, 0, 0)), rectLayout, this.m_treeTextFormat);
			}
		}


		/// <summary>
		/// Expand/collapse only when the expand button is clicked
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void treeMain_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
		{
			if (e.Node != null)
			{
				int totalBarWidth = s_BarWidthPercentage * this.treeMain.Bounds.Width / 100;	
				int expandButtonEnd = totalBarWidth + (e.Node.Level * treeMain.Indent) + 16;

				if (e.Location.X > totalBarWidth && e.Location.X < expandButtonEnd)
				{
					if (e.Node.IsExpanded)
						e.Node.Collapse();
					else
						e.Node.Expand();
				}
			}

		}


		/// <summary>
		/// Draw a bar
		/// </summary>
		/// <param name="e"></param>
		/// <param name="x"></param>
		/// <param name="y"></param>
		/// <param name="w"></param>
		/// <param name="h"></param>
		/// <param name="percentage"></param>
		private void drawBar(DrawTreeNodeEventArgs e, int x, int y, int w, int h, double percentage1, double percentage2, Color colMain, Color colSecondary)
		{
			Rectangle rect;
			// First percentage is 
			double filledWidth1 = percentage1 * ((double) w);
			rect = new Rectangle(x + 1, y + 1, (int)filledWidth1, h - 1);
			e.Graphics.FillRectangle(new SolidBrush(colMain), rect);
			if (percentage2 != 0) // Percentage 2 is total
			{
				double filledWidth2 = percentage2 * ((double)w);
				rect = new Rectangle(x + 1, y + 1, (int)filledWidth2, h - 1);
				e.Graphics.FillRectangle(new SolidBrush(colSecondary), rect);
			}

			Pen penBorder = null;
			if (percentage1 == 0)
				penBorder = new Pen(m_barBorderColorVague, 1);
			else
				penBorder = new Pen(m_barBorderColor, 1);
			rect = new Rectangle(x, y, w, h);
			e.Graphics.DrawRectangle(penBorder, rect);
		}


		/// <summary>
		/// Draw tick marks for the logarithmic bar
		/// </summary>
		/// <param name="e"></param>
		/// <param name="x"></param>
		/// <param name="y"></param>
		/// <param name="w"></param>
		/// <param name="h"></param>
		/// <param name="rootFolderSize"></param>
		private void drawLogMarks(DrawTreeNodeEventArgs e, int x, int y, int w, int h, ulong rootFolderSize)
		{
			double divider = TreeBuilder.KILOBYTE;
			double totalSize = Math.Log10(((double)rootFolderSize) / divider);
			if (totalSize == 0) totalSize = 1;

			Pen penBorderV = new Pen(m_barBorderColorVague, 1);
			Pen penBorder = new Pen(m_barBorderColor, 1);
			ulong[] markSizes = { TreeBuilder.KILOBYTE10, TreeBuilder.KILOBYTE100, TreeBuilder.MEGABYTE, TreeBuilder.MEGABYTE10, TreeBuilder.MEGABYTE100, TreeBuilder.GIGABYTE, TreeBuilder.GIGABYTE10, TreeBuilder.GIGABYTE100, TreeBuilder.TERABYTE, TreeBuilder.TERABYTE10, TreeBuilder.TERABYTE100, TreeBuilder.PETABYTE };
			bool[]  markHigh =  { false,   false,    true,	 false,   false,    true,   false,   false,	   true,   false,   false,    true };
			for(int n = 0; n < markSizes.Length; n++)
			{
				if (markSizes[n] < rootFolderSize)
				{
					double dist = Math.Log10(((double)markSizes[n]) / divider) / totalSize;
					dist = (dist * w) + x;
					if (markHigh[n])
						e.Graphics.DrawLine(penBorder, (int)dist, y + 1, (int)dist, y + h - 1);
					else
						e.Graphics.DrawLine(penBorderV, (int)dist, y + 1, (int)dist, y + h - 1);
				}
			}
		}


		/// <summary>
		/// A folder is selected
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void treeMain_AfterSelect(object sender, TreeViewEventArgs e)
		{
			// Store the current selected path
			this.m_currentSelectedFolder = e.Node.Tag as Folder;
			this.m_currentSelectedPath = (m_currentSelectedFolder != null) ? m_currentSelectedFolder.FullPath : "";

			// Change the listfiles
			this.lvwTreeFiles.ChangeSelectedFolder(e.Node);
		}

		
		/// <summary>
		/// Form now closing
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
		{
			this.m_isClosing = true;
		}


		/// <summary>
		/// Change the current viewing mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="newViewMode"></param>
		private void ChangeViewingMode(ViewOptions newViewMode)
		{
			tsbPercentage.Checked = (newViewMode == ViewOptions.Percentage);
			tsbRelativeToParent.Checked = (newViewMode == ViewOptions.PercentageRelativeToParent);
			tsbPercentageLog.Checked = (newViewMode == ViewOptions.PercentageLog);

			tsmiPercentage.Checked = (newViewMode == ViewOptions.Percentage);
			tsmiRelative.Checked = (newViewMode == ViewOptions.PercentageRelativeToParent);
			tsmiLogarithmic.Checked = (newViewMode == ViewOptions.PercentageLog);

			// Store the last selected view mode
			DiskSpaceAnalyzer.Properties.Settings.Default.Main_ViewOptions = newViewMode.ToString();
			DiskSpaceAnalyzer.Properties.Settings.Default.Save();

			m_currentViewOptions = newViewMode;
			treeMain.Invalidate();
		}

		/// <summary>
		/// Change the current viewing mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="newViewMode"></param>
		private void ChangeViewingModeEx(ViewOptionsEx newViewModeEx)
		{
			tsbSize.Checked = (newViewModeEx == ViewOptionsEx.Size);
			tsbNumber.Checked = (newViewModeEx == ViewOptionsEx.NumberOfFiles);

			tsmiSize.Checked = (newViewModeEx == ViewOptionsEx.Size);
			tsmiNumber.Checked = (newViewModeEx == ViewOptionsEx.NumberOfFiles);

			// Store the last selected view mode
			DiskSpaceAnalyzer.Properties.Settings.Default.Main_ViewOptionsEx = newViewModeEx.ToString();
			DiskSpaceAnalyzer.Properties.Settings.Default.Save();

			m_currentViewOptionsEx = newViewModeEx;
			treeMain.Invalidate();
		}

		private void tsbPercentage_Click(object sender, EventArgs e) { ChangeViewingMode(ViewOptions.Percentage); }
		private void tsbRelativeToParent_Click(object sender, EventArgs e) { ChangeViewingMode(ViewOptions.PercentageRelativeToParent); }
		private void tsbPercentageLog_Click(object sender, EventArgs e)	{ ChangeViewingMode(ViewOptions.PercentageLog); }
		private void tsmiPercentage_Click(object sender, EventArgs e) { ChangeViewingMode(ViewOptions.Percentage); }
		private void tsmiRelative_Click(object sender, EventArgs e) { ChangeViewingMode(ViewOptions.PercentageRelativeToParent); }
		private void tsmiLogarithmic_Click(object sender, EventArgs e) { ChangeViewingMode(ViewOptions.PercentageLog); }

		private void tsbSize_Click(object sender, EventArgs e) { ChangeViewingModeEx(ViewOptionsEx.Size); }
		private void tsbNumber_Click(object sender, EventArgs e) { ChangeViewingModeEx(ViewOptionsEx.NumberOfFiles); }
		private void tsmiSize_Click(object sender, EventArgs e) { ChangeViewingModeEx(ViewOptionsEx.Size); }
		private void tsmiNumber_Click(object sender, EventArgs e) { ChangeViewingModeEx(ViewOptionsEx.NumberOfFiles); }

		/// <summary>
		/// Refresh button clicked, show startup screen again
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tsbRefresh_Click(object sender, EventArgs e) { this.ShowScanDialog(); }
		private void tsmiRefresh_Click(object sender, EventArgs e) { this.ShowScanDialog(); }



		/// <summary>
		/// Open the current selected folder in internet explorer
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void btnGo_Click(object sender, EventArgs e)
		{
			Process.Start(this.m_currentSelectedPath);
		}



		/// <summary>
		/// Show the about screen
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
		{
			AboutForm formAbout = new AboutForm();
			formAbout.ShowDialog();			
		}


		/// <summary>
		/// Show the settings screen
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void settingsToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SettingsForm formSettings = new SettingsForm();
			if (formSettings.ShowDialog() == System.Windows.Forms.DialogResult.OK)
			{
				// Refresh settings
				LoadUserSettings();

				// Link the top list
				this.lvwTopFiles.Initialize(DiskSpaceAnalyzer.Properties.Settings.Default.Main_TopListNumber);

				// Refresh the tree
				FillTree();
			}
		}

		
		/// <summary>
		/// Load all (relevant) user settings
		/// </summary>
		private void LoadUserSettings()
		{
			this.m_barColorMain = DiskSpaceAnalyzer.Properties.Settings.Default.Bar_Color1;
			this.m_barColorSecondary = DiskSpaceAnalyzer.Properties.Settings.Default.Bar_Color2;
			this.m_barColorRoot = DiskSpaceAnalyzer.Properties.Settings.Default.Bar_Root;
			this.m_currentUnit = (UnitTypes)DiskSpaceAnalyzer.Properties.Settings.Default.Main_Unit;
			
			this.lvwTreeFiles.GridLines = DiskSpaceAnalyzer.Properties.Settings.Default.ShowGridLines;
			this.lvwTopFiles.GridLines = DiskSpaceAnalyzer.Properties.Settings.Default.ShowGridLines;
			this.lvwSearchResults.GridLines = DiskSpaceAnalyzer.Properties.Settings.Default.ShowGridLines;

			this.lvwTreeFiles.ChangeUnit(this.m_currentUnit);
			this.lvwTopFiles.ChangeUnit(this.m_currentUnit);
			this.lvwSearchResults.ChangeUnit(this.m_currentUnit);
		}


		/// <summary>
		/// Change the viewing mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void cmbOptions_SelectedIndexChanged(object sender, EventArgs e)
		{
			FillTree();

			// Change the viewmode inside the listfiles
			TreeViewMode newViewMode = (((KeyValuePair<TreeViewMode, string>)cmbOptions.SelectedItem).Key);
			this.lvwTreeFiles.ChangeViewMode(newViewMode);
			this.lvwTopFiles.ChangeViewMode(newViewMode);
			this.lvwSearchResults.ChangeViewMode(newViewMode);
		}


		/// <summary>
		/// Put all folders in the tree
		/// </summary>
		private void FillTree()
		{
			TreeViewMode viewMode = ((KeyValuePair<TreeViewMode, string>)cmbOptions.SelectedItem).Key;
			
			// Store this last selected viewing mode
			DiskSpaceAnalyzer.Properties.Settings.Default.Main_ViewMode = cmbOptions.SelectedIndex;
			DiskSpaceAnalyzer.Properties.Settings.Default.Save();


			this.treeMain.BeginUpdate();

			// Clear the toplists
			this.lvwTopFiles.ClearList();

			// Refill the tree
			this.treeMain.Enabled = false;
			foreach (TreeNode node in this.treeMain.Nodes)
			{
				RootFolder rootFolder = node.Tag as RootFolder;
				if (rootFolder != null && rootFolder.Parser != null)
				{
					// Create the builder => this will fill the entire tree!!
					rootFolder.Builder = new TreeBuilder(node, rootFolder, viewMode, m_currentSelectedFolder);

					// Add the rootfolder to the toplist
					this.lvwTopFiles.AddRoot(rootFolder);
				}
			}
			this.treeMain.Enabled = true;

			// Update the top list
			this.lvwTopFiles.RefreshFileList();

			this.treeMain.EndUpdate();
		}


		/// <summary>
		/// Show the properties of a file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tsmiFolderProperties_Click(object sender, EventArgs e)
		{
			WindowsDialogs.ShowFileProperties(this.m_currentSelectedPath);
		}

		/// <summary>
		/// Expand or collapse the selected folder
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tsmiFolderExpand_Click(object sender, EventArgs e)
		{
			if (this.treeMain.SelectedNode != null)
			{
				if (this.treeMain.SelectedNode.IsExpanded)
					this.treeMain.SelectedNode.Collapse();
				else
					this.treeMain.SelectedNode.Expand();
			}
		}

		/// <summary>
		/// Expand all folders
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tsmiFolderExpandAll_Click(object sender, EventArgs e)
		{
			if (this.treeMain.SelectedNode != null)
				this.treeMain.SelectedNode.ExpandAll();
		}

		/// <summary>
		/// Open an explorer
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tsmiFolderOpenExplorer_Click(object sender, EventArgs e)
		{
			Process.Start(this.m_currentSelectedPath);
		}

		/// <summary>
		/// Open command window here
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tsmiFolderOpenCommand_Click(object sender, EventArgs e)
		{
			string drive = Path.GetPathRoot(this.m_currentSelectedPath).Trim('\\', '/');
			ProcessStartInfo start = new ProcessStartInfo("cmd.exe", string.Format("/K \"{0} & cd {1}\"", drive, this.m_currentSelectedPath) );
			Process.Start(start);
		}


		/// <summary>
		/// Open the properties of the selected file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tsmiFileProperties_Click(object sender, EventArgs e)
		{
			Myriadbits.NTFS.File fd = this.CurrentSelectedFile;
			if (fd != null) // Only for files
				WindowsDialogs.ShowFileProperties(fd.FullPath);
		}

		
		/// <summary>
		/// Delete the selected file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tsmiFileDelete_Click(object sender, EventArgs e)
		{
			Myriadbits.NTFS.File fd = this.CurrentSelectedFile;
			if (fd != null) // Only for files
			{
				if (MessageBox.Show(string.Format("Are you absolutely sure you want to delete the following file?\n{0}", fd.FullPath), "Confirm delete", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question) == System.Windows.Forms.DialogResult.Yes)
				{
					try
					{
						// Stop the check-free space timer
						this.timerCheckFree.Stop();

						// Really delete the file
						System.IO.File.Delete(fd.FullPath);

						// Now do a trick, remove the file reference number from the folder
						fd.Folder.RemoveFile(fd.ReferenceNumber);

						// Set the number of files that should be shown in the virtual list
						this.lvwTreeFiles.DeleteFile();

						// Re-calculate the largest files if this was a file with a ranking
						if (fd.Rank != -1)
						{
							fd.Folder.Parser.CalculateLargestFiles();

							// And refresh the top list
							this.lvwTopFiles.RefreshFileList();
						}

						// Restart the timer
						this.m_fDeletingFile = true; // Set flag to indicate that the timer should not check the free space 
						this.timerCheckFree.Start();
					}
					catch(Exception ex)
					{
						MessageBox.Show(string.Format("Failed to delete {0}.\nError: {1}", fd.FullPath, ex.Message), "Error", MessageBoxButtons.OKCancel, MessageBoxIcon.Error);
					}
				}
			}
		}


		/// <summary>
		/// (Try to) Open the selected file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tsmiFileOpen_Click(object sender, EventArgs e)
		{
			Myriadbits.NTFS.File fd = this.CurrentSelectedFile;
			if (fd != null) // Only for files
			{
				try
				{
					Process.Start(fd.FullPath);
				}
				catch (Exception ex)
				{
					MessageBox.Show(string.Format("Failed to start {0}.\nError: {1}", fd.FullPath, ex.Message), "Error", MessageBoxButtons.OKCancel, MessageBoxIcon.Error);
				}
			}
		}


		/// <summary>
		/// Returns the full path to the current selected file
		/// </summary>
		public Myriadbits.NTFS.File CurrentSelectedFile
		{
			get
			{
				if (this.tabMain.SelectedIndex == 0)
				{
					FileData fd = this.lvwTreeFiles.SelectedFileData;
					if (fd != null && !fd.IsFolder) // Only for files
						return fd.File;
				}
				else
					return m_currentSelectedFile;
				return null;
			}
		}


		/// <summary>
		/// Double click on item
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void listFiles_DoubleClick(object sender, EventArgs e)
		{
			FileData fd = this.lvwTreeFiles.SelectedFileData;
			if (fd != null)
			{
				if (fd.IsFolder)
				{
					TreeNode tn = this.treeMain.SelectedNode;
					if (!tn.IsExpanded)
						tn.Expand();
					foreach (TreeNode subnode in tn.Nodes)
					{
						Folder subFolder = subnode.Tag as Folder;
						if (subFolder != null && subFolder == fd.ThisFolder)
						{
							this.treeMain.SelectedNode = subnode;
							break;
						}
					}					
				}
				else
				{
					// Open file
					try
					{
						System.Diagnostics.Process.Start(fd.FullPath);
					}
					catch (Exception ex)
					{
						MessageBox.Show(string.Format("Failed to execute {0}.\nError: {1}", fd.FullPath, ex.Message), "Error", MessageBoxButtons.OKCancel, MessageBoxIcon.Error);
					}
				}
			}
		}


		/// <summary>
		/// Folder menu opening, modify text
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void cmsFolders_Opening(object sender, System.ComponentModel.CancelEventArgs e)
		{
			if (this.treeMain.SelectedNode != null)
			{
				if (this.treeMain.SelectedNode.IsExpanded)
					tsmiFolderExpand.Text = "Collapse";
				else
					tsmiFolderExpand.Text = "Expand";
			}
		}


		/// <summary>
		/// Check if the free space has changed
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void timerCheckFree_Tick(object sender, EventArgs e)
		{
			DriveInfo[] allDrives = DriveInfo.GetDrives();
			long margin = (long) TreeBuilder.MEGABYTE10;
			for (int n = 0; n < allDrives.Count() && n < this.m_vecFreeSpacesPerDrive.Count(); n++)
			{
				// Deleting a file, just store the current free space
				if (this.m_fDeletingFile)
				{					
					this.m_vecFreeSpacesPerDrive[n] = allDrives[n].AvailableFreeSpace;
					this.m_fDeletingFile = false;
				}

				if (allDrives[n].AvailableFreeSpace <= (this.m_vecFreeSpacesPerDrive[n] - margin) ||
					allDrives[n].AvailableFreeSpace >= (this.m_vecFreeSpacesPerDrive[n] + margin))
				{
					// Free space has changed a lot, request refresh
					if (!m_fShowingRefreshQuestion)
					{
						this.m_fShowingRefreshQuestion = true;
						if (MessageBox.Show(this, string.Format("It seems that the free space on drive {0} changed from {1} to {2}.\nDo you want to refresh the tree?", allDrives[n].Name, m_currentUnit.GetUnitText(allDrives[n].AvailableFreeSpace), m_currentUnit.GetUnitText(this.m_vecFreeSpacesPerDrive[n])), "Refresh", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question) == System.Windows.Forms.DialogResult.Yes)
						{
							ShowScanDialog(true); // Start scanning direct
							this.m_fShowingRefreshQuestion = false;
						}
					}
					this.m_vecFreeSpacesPerDrive[n] = allDrives[n].AvailableFreeSpace;
				}				
			}
		}


		/// <summary>
		/// Selection in top files changed, make sure the same folder is selected in the main tree
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void lvwTopFiles_SelectedIndexChanged(object sender, EventArgs e)
		{
			SelectFileInTree(this.lvwTopFiles.SelectedFolder, this.lvwTopFiles.SelectedFile);
		}


		/// <summary>
		/// Selection in search files changed, make sure the same folder is selected in the main tree
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void lvwSearchResults_SelectedIndexChanged(object sender, EventArgs e)
		{
			SelectFileInTree(this.lvwSearchResults.SelectedFolder, this.lvwSearchResults.SelectedFile);
		}


		/// <summary>
		/// Select a file in the tree
		/// </summary>
		/// <param name="folderToSelect"></param>
		/// <param name="fileToSelect"></param>
		private void SelectFileInTree(Folder folderToSelect, Myriadbits.NTFS.File fileToSelect)
		{
			if (folderToSelect == null)
				return;

			m_currentSelectedFolder = folderToSelect;
			m_currentSelectedPath = (folderToSelect != null) ? folderToSelect.FullPath : "";
			m_currentSelectedFile = fileToSelect;

			TreeNode nodeToSelect = null;
			foreach (TreeNode tn in this.treeMain.Nodes)
			{
				nodeToSelect = FindFolderInTree(tn, folderToSelect.ReferenceNumber, folderToSelect.Name);
				if (nodeToSelect != null)
					break;
			}
			if (nodeToSelect != null)
			{
				this.treeMain.SelectedNode = nodeToSelect;

				if (fileToSelect == null)
					return;

				// Select this file in the file list
				this.lvwTreeFiles.SelectFile(fileToSelect);
			}
		}


		/// <summary>
		/// Doubleclick on search results = jump to tree
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void lvwSearchResults_DoubleClick(object sender, EventArgs e)
		{
			this.tabMain.SelectedTab = this.tabPageAll;
		}


		/// <summary>
		/// Find a folder in a tree (recursive!)
		/// </summary>
		/// <param name="tn"></param>
		/// <param name="folderToSelect"></param>
		/// <returns></returns>
		private TreeNode FindFolderInTree(TreeNode tn, ulong folderToSelect, string folderToSelectName)
		{
			Folder currentFolder = tn.Tag as Folder;
			if (currentFolder.ReferenceNumber == folderToSelect && currentFolder.Name == folderToSelectName)
				return tn;

			foreach (TreeNode child in tn.Nodes)
			{
				TreeNode retVal = FindFolderInTree(child, folderToSelect, folderToSelectName);
				if (retVal != null)
					return retVal;
			}
			return null;
		}


		/// <summary>
		/// Double click on an top list item = jump to tree
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void lvwTopFiles_DoubleClick(object sender, EventArgs e)
		{
			this.tabMain.SelectedTab = this.tabPageAll;
		}


		/// <summary>
		/// When the enter key is pressed, start searching
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void txtSearch_KeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Return)
				btnSearch_Click(null, null);
		}


		/// <summary>
		/// Search files
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void btnSearch_Click(object sender, EventArgs e)
		{
			this.txtSearch.Enabled = false;
			this.lvwSearchResults.Enabled = false;
			this.btnSearch.Enabled = false;

			// TODO Might be better to do this in a seperate thread...
			string searchMask = string.Format("*{0}*", this.txtSearch.Text);
			Stopwatch sw = this.lvwSearchResults.SearchFiles(searchMask, this.treeMain);
			sw.Stop();

			if (this.lvwSearchResults.CountShown != this.lvwSearchResults.CountFound)
				this.txtSearchInfo.Text = string.Format("Search found {0} files, but only showing the first {1} (in {2} ms)", this.lvwSearchResults.CountFound, this.lvwSearchResults.CountShown, sw.ElapsedMilliseconds);
			else
				this.txtSearchInfo.Text = string.Format("Search found {0} files (in {1} ms)", this.lvwSearchResults.CountFound, sw.ElapsedMilliseconds);
			this.txtSearch.Enabled = true;
			this.lvwSearchResults.Enabled = true;
			this.btnSearch.Enabled = true;
		}		

	}


	public enum ViewOptions
	{
		Percentage,
		PercentageLog,
		PercentageRelativeToParent
	};

	public enum ViewOptionsEx
	{
		Size,
		NumberOfFiles,
		FileID
	};

	
	

}
