﻿using Myriadbits.NTFS;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Myriadbits.DiskSpaceAnalyzer
{
	public class SearchFileList : ListView
	{
		private Color				m_textColor = MainForm.m_textColor;
		private Color				m_textColorZeroSize = MainForm.m_textColorZeroSize;
		private TreeViewMode		m_treeViewMode = TreeViewMode.AllFiles;
		private UnitTypes			m_currentUnit = UnitTypes.Auto;
		private int					m_numberOfFilesFound = 0;
		private ImageList			m_imageFiles = new ImageList();
		private List<KeyValuePair<File, ListViewItem>> m_allFiles = new List<KeyValuePair<File, ListViewItem>>();
		
		public int SortColumn { get; set; }
		public bool SortAscending { get; set; }

		/// <summary>
		/// Constructor
		/// </summary>
		public SearchFileList()
		{
			this.SortColumn = 0;
			this.SortAscending = true;			
			this.ColumnClick += this.onColumnClicked;
			this.m_imageFiles.ColorDepth = ColorDepth.Depth32Bit;
			this.SmallImageList = this.m_imageFiles;
		}


		/// <summary>
		/// Change the current view mode
		/// </summary>
		/// <param name="newViewMode"></param>
		public void ChangeViewMode(TreeViewMode newViewMode)
		{
			if (newViewMode != this.m_treeViewMode)
			{
				this.m_treeViewMode = newViewMode;
				RefreshFileList();
			}
		}


		/// <summary>
		/// The current unit type has changed (bytes to MB or GB, or...)
		/// </summary>
		/// <param name="newUnitType"></param>
		public void ChangeUnit(UnitTypes newUnitType)
		{
			if (newUnitType != m_currentUnit)
			{
				m_currentUnit = newUnitType;
				RefreshFileList();
			}
		}

		/// <summary>
		/// Search for all files
		/// </summary>
		/// <param name="searchMask"></param>
		/// <param name="treeMain"></param>
		public Stopwatch SearchFiles(string searchMask, TreeView treeMain)
		{
			List<Myriadbits.NTFS.File> allFiles = new List<Myriadbits.NTFS.File>();

			Stopwatch sw = Stopwatch.StartNew();
			this.m_allFiles.Clear();
			foreach (TreeNode node in treeMain.Nodes)
			{
				RootFolder rootFolder = node.Tag as RootFolder;
				if (rootFolder != null && rootFolder.Parser != null)
				{
					this.m_allFiles.AddRange(rootFolder.Parser.FindFiles(searchMask).Select(a => new KeyValuePair<File, ListViewItem>(a, new ListViewItem())));
				}
			}
			this.Items.Clear();
			sw.Stop();

			// Limit the number of files
			m_numberOfFilesFound = this.m_allFiles.Count;
			if (m_numberOfFilesFound > 10000)
			{
				MessageBox.Show(string.Format("Found more then {0} files that match the search criteria.\nOnly showing the first 10000 files!", this.m_allFiles.Count), "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
				this.m_allFiles = this.m_allFiles.Take(10000).ToList();
			}


			// Clear all images (just to be sure we don't waste memory...)
			m_imageFiles.Images.Clear();
			foreach (KeyValuePair<File, ListViewItem> kvp in this.m_allFiles)
			{
				// Get the icon type of this folder
				string iconType = FileIcon.GetIconType(kvp.Key.FullPath);

				// Add the icon to the imagelist (If not already present)		
				if (!string.IsNullOrEmpty(iconType))
				{
					if (!m_imageFiles.Images.ContainsKey(iconType))
					{
						// Load the icon type
						Icon iconNew = FileIcon.GetSmallIcon(kvp.Key.FullPath);
						if (iconNew != null)
							m_imageFiles.Images.Add(iconType, iconNew);
					}

					kvp.Value.Text = kvp.Key.Name;
					kvp.Value.ImageKey = iconType;
					kvp.Value.SubItems.Add(m_currentUnit.GetUnitText(kvp.Key.Size));
					kvp.Value.SubItems.Add(kvp.Key.Folder.FullPath);
					kvp.Value.Tag = kvp.Key;
				}
			}

			// Update display
			RefreshFileList();

			return sw;
		}


		/// <summary>
		/// Return the number of items that are found
		/// </summary>
		public int CountFound
		{
			get
			{
				return m_numberOfFilesFound;
			}
		}


		/// <summary>
		/// Return the number of items that are shown
		/// </summary>
		public int CountShown
		{
			get
			{
				return m_numberOfFilesFound;
			}
		}


		/// <summary>
		/// Header click -> sort the list
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void onColumnClicked(object sender, ColumnClickEventArgs e)
		{
			if (this.SortColumn == e.Column)
			{
				this.SortAscending = !this.SortAscending;
			}
			else
			{
				this.SortAscending = true;
				this.SortColumn = e.Column;
			}

			// Refresh/update the list
			RefreshFileList();
		}

		
		/// <summary>
		/// Fill the list with data
		/// </summary>
		public void RefreshFileList()
		{
			// Clear whole list
			this.Items.Clear();

			switch (this.SortColumn)
			{
				case 0: // File names
					if (this.SortAscending)
						this.m_allFiles = m_allFiles.OrderBy(a => a.Key.Name).ToList();
					else
						this.m_allFiles = m_allFiles.OrderByDescending(a => a.Key.Name).ToList();
					break;
				case 1: // Size of files
					if (this.SortAscending)
						this.m_allFiles = m_allFiles.OrderByDescending(a => a.Key.Size).ToList();
					else
						this.m_allFiles = m_allFiles.OrderBy(a => a.Key.Size).ToList();
					break;
				case 2: // Path
					if (this.SortAscending)
						this.m_allFiles = m_allFiles.OrderBy(a => a.Key.Folder.FullPath).ToList();
					else
						this.m_allFiles = m_allFiles.OrderByDescending(a => a.Key.Folder.FullPath).ToList();
					break;
			}

			Stopwatch swsort = Stopwatch.StartNew();

			// Now add allfiles to the list
			foreach (KeyValuePair<File, ListViewItem> kvp in this.m_allFiles)
			{
				kvp.Value.SubItems[1].Text = m_currentUnit.GetUnitText(kvp.Key.Size); // Unit could be changed, set it again

				if (!TreeBuilder.IsFileVisible(kvp.Key, this.m_treeViewMode))
					kvp.Value.ForeColor = this.m_textColorZeroSize;
				else
					kvp.Value.ForeColor = this.m_textColor;

				this.Items.Add(kvp.Value);
			}

			swsort.Stop();
		}


		/// <summary>
		/// Get the current selected folder
		/// </summary>
		/// <returns></returns>
		public Folder SelectedFolder
		{
			get
			{
				if (this.SelectedItems.Count > 0)
				{
					File file = this.SelectedItems[0].Tag as File;
					if (file != null)
						return file.Folder;
				}
				return null;
			}
		}


		/// <summary>
		/// Get the current selected file
		/// </summary>
		/// <returns></returns>
		public File SelectedFile
		{
			get
			{
				if (this.SelectedItems.Count > 0)
					return this.SelectedItems[0].Tag as File;
				return null;
			}
		}


	}
}
