﻿//
// Myriadbits Disk space analyzer 
// Find free space on your NTFS disk
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Myriadbits.NTFS;

namespace Myriadbits.DiskSpaceAnalyzer
{
	public enum TreeViewMode
	{
		AllFiles = 0,
		HideEmptyFolders,
		FolderLargerThen1MB,
		FolderLargerThen10MB,
		FolderLargerThen100MB,
		FolderLargerThen1GB,
		FolderLargerThen10GB,
		FolderLargerThen100GB,
		FileLargerThen1MB,
		FileLargerThen10MB,
		FileLargerThen100MB,
		FileLargerThen1GB,
		FileLargerThen10GB,
		FileLargerThen100GB,
		FileLargest10,
		FileLargest100,
		FileLargest1000,
	};


	/// <summary>
	/// Helper class to build the treeview when give a folder
	/// </summary>
	public class TreeBuilder
	{
		public static ulong KILOBYTE = 1024;
		public static ulong KILOBYTE10 = 10 * KILOBYTE;
		public static ulong KILOBYTE100 = 100 * KILOBYTE;
		public static ulong MEGABYTE = 1024 * KILOBYTE;
		public static ulong MEGABYTE10 = 10 * MEGABYTE;
		public static ulong MEGABYTE100 = 100 * MEGABYTE;
		public static ulong GIGABYTE = (ulong)1024 * MEGABYTE;
		public static ulong GIGABYTE10 = (ulong)10 * GIGABYTE;
		public static ulong GIGABYTE100 = (ulong)100 * GIGABYTE;
		public static ulong TERABYTE = (ulong)1024 * GIGABYTE;
		public static ulong TERABYTE10 = 10 * TERABYTE;
		public static ulong TERABYTE100 = 100 * TERABYTE;
		public static ulong PETABYTE = 1024 * TERABYTE;

		private List<ulong> listLargestFiles = null;

		/// <summary>
		/// Build the tree
		/// </summary>
		/// <param name="treeMain"></param>
		/// <param name="folders"></param>
		public TreeBuilder(TreeNode treeNode, RootFolder rootFolder, TreeViewMode viewMode, Folder selectedFolder)
		{
			if (treeNode == null) return;

			// Get the largest X files from the rootfolder if the viewmode is largest X
			if (viewMode == TreeViewMode.FileLargest10)
				this.listLargestFiles = rootFolder.Parser.LargestFiles.Take(10).Select(a => a.Key).ToList();
			else if (viewMode == TreeViewMode.FileLargest100)
				this.listLargestFiles = rootFolder.Parser.LargestFiles.Take(100).Select(a => a.Key).ToList();
			else if (viewMode == TreeViewMode.FileLargest1000)
				this.listLargestFiles = rootFolder.Parser.LargestFiles.Take(1000).Select(a => a.Key).ToList();
			else
				this.listLargestFiles = rootFolder.Parser.LargestFiles.Select(a => a.Key).ToList();

			bool autoExpand = false;
			switch (viewMode)
			{
				case TreeViewMode.FileLargest10:
				case TreeViewMode.FileLargest100:
					autoExpand = true;
					break;

				case TreeViewMode.FileLargerThen100MB:
				case TreeViewMode.FileLargerThen1GB:
				case TreeViewMode.FileLargerThen10GB:
				case TreeViewMode.FileLargerThen100GB:
					autoExpand = true;
					break;
			}

			// Clear all subnodes
			treeNode.Nodes.Clear();
			//string name = string.Format("{0} [{1} files, {2} bytes]", rootFolder.Name, rootFolder.NumberOfFiles, rootFolder.Size);
			TreeNode selectedNode = null;
			if (rootFolder != null)
			{
				string name = rootFolder.Name;
				treeNode.Tag = rootFolder;
				selectedNode = AddChildren(treeNode, rootFolder, viewMode, selectedFolder, autoExpand);			
				treeNode.Expand();
			}

			if (selectedNode != null)
			{
				treeNode.TreeView.SelectedNode = selectedNode;
			}
		}


		/// <summary>
		/// Add all children of a parent
		/// </summary>
		/// <param name="parentNode"></param>
		/// <param name="parentId"></param>
		private TreeNode AddChildren(TreeNode parentNode, Folder folder, TreeViewMode viewMode, Folder selectedFolder, bool expand)
		{
			TreeNode retVal = null;
			if (folder == selectedFolder)
			{
				// Yes, found!
				retVal = parentNode;
			}

			if (folder.HasChildren)
			{
				foreach (Folder child in folder.Children.OrderBy(a => a.Name))
				{
					if (IsChildVisible(child, viewMode))
					{
						TreeNode tn = parentNode.Nodes.Add(child.ReferenceNumber.ToString(), child.Name, 1, 1);
						tn.Tag = child;
						TreeNode childSelected = AddChildren(tn, child, viewMode, selectedFolder, expand);
						if (childSelected != null)
							retVal = childSelected;

						if (expand)
							parentNode.Expand();
					}
				}
			}
			return retVal;
		}


		/// <summary>
		/// Add all children of a parent
		/// </summary>
		/// <param name="parentNode"></param>
		/// <param name="parentId"></param>
		public ulong CountVisibleChildren(Folder folder, TreeViewMode viewMode)
		{
			ulong counter = 0;
			if (folder.HasChildren)
			{
				foreach (Folder child in folder.Children.OrderBy(a => a.Name))
				{
					if (this.IsChildVisible(child, viewMode))
					{
						counter++;
					}
				}
			}
			return counter;
		}


		/// <summary>
		/// Returns true if the child should be visible
		/// </summary>
		/// <param name="child"></param>
		/// <param name="viewMode"></param>
		/// <returns></returns>
		public bool IsChildVisible(Folder child, TreeViewMode viewMode)
		{
			switch (viewMode)
			{
				case TreeViewMode.AllFiles:
					return true;

				//
				// Show all folders who's size is at least X
				case TreeViewMode.HideEmptyFolders:
					return (child.Size > 0);
				case TreeViewMode.FolderLargerThen1MB:
					return (child.Size >= TreeBuilder.MEGABYTE);
				case TreeViewMode.FolderLargerThen10MB:
					return (child.Size >= TreeBuilder.MEGABYTE10);
				case TreeViewMode.FolderLargerThen100MB:
					return (child.Size >= TreeBuilder.MEGABYTE100);
				case TreeViewMode.FolderLargerThen1GB:
					return (child.Size >= TreeBuilder.GIGABYTE);
				case TreeViewMode.FolderLargerThen10GB:
					return (child.Size >= TreeBuilder.GIGABYTE10);
				case TreeViewMode.FolderLargerThen100GB:
					return (child.Size >= TreeBuilder.GIGABYTE100);

				//
				// Show all folders that holds at least one file larger then X
				// OR 1 of its childs hold a file larger then X
				case TreeViewMode.FileLargerThen1MB:
					return child.HasFileLargerThenSize(TreeBuilder.MEGABYTE);
				case TreeViewMode.FileLargerThen10MB:
					return child.HasFileLargerThenSize(TreeBuilder.MEGABYTE10);
				case TreeViewMode.FileLargerThen100MB:
					return child.HasFileLargerThenSize(TreeBuilder.MEGABYTE100);
				case TreeViewMode.FileLargerThen1GB:
					return child.HasFileLargerThenSize(TreeBuilder.GIGABYTE);
				case TreeViewMode.FileLargerThen10GB:
					return child.HasFileLargerThenSize(TreeBuilder.GIGABYTE10);
				case TreeViewMode.FileLargerThen100GB:
					return child.HasFileLargerThenSize(TreeBuilder.GIGABYTE100);

				//
				// Show all folders that hold one of the files in the top X
				// Or the parent folders of that folders
				case TreeViewMode.FileLargest10:
				case TreeViewMode.FileLargest100:
				case TreeViewMode.FileLargest1000:
					return child.HasFileFromCollection(this.listLargestFiles);
			}
			return false;
		}


		/// <summary>
		/// Returns true if the child should be visible
		/// </summary>
		/// <param name="child"></param>
		/// <param name="viewMode"></param>
		/// <returns></returns>
		public bool IsFileVisible(FileData file, TreeViewMode viewMode)
		{
			return IsFileVisible(file.File, viewMode);
		}


		/// <summary>
		/// Static method to determine if a file should be shown in a viewmode
		/// </summary>
		/// <returns></returns>
		public static bool IsFileVisible(File file, TreeViewMode viewMode)
		{
			switch (viewMode)
			{
				case TreeViewMode.AllFiles:
					return true;

				//
				// Show all folders who's size is at least X
				case TreeViewMode.HideEmptyFolders:
				case TreeViewMode.FolderLargerThen1MB:
				case TreeViewMode.FolderLargerThen10MB:
				case TreeViewMode.FolderLargerThen100MB:
				case TreeViewMode.FolderLargerThen1GB:
				case TreeViewMode.FolderLargerThen10GB:
				case TreeViewMode.FolderLargerThen100GB:
					return true; // This is a file, not a folder, allow all

				//
				// Show all folders that holds at least one file larger then X
				// OR 1 of its childs hold a file larger then X
				case TreeViewMode.FileLargerThen1MB:
					return file.Size >= TreeBuilder.MEGABYTE;
				case TreeViewMode.FileLargerThen10MB:
					return file.Size >= TreeBuilder.MEGABYTE10;
				case TreeViewMode.FileLargerThen100MB:
					return file.Size >= TreeBuilder.MEGABYTE100;
				case TreeViewMode.FileLargerThen1GB:
					return file.Size >= TreeBuilder.GIGABYTE;
				case TreeViewMode.FileLargerThen10GB:
					return file.Size >= TreeBuilder.GIGABYTE10;
				case TreeViewMode.FileLargerThen100GB:
					return file.Size >= TreeBuilder.GIGABYTE100;

				//
				// Show all folders that hold one of the files in the top X
				// Or the parent folders of that folders
				case TreeViewMode.FileLargest10:
				case TreeViewMode.FileLargest100:
				case TreeViewMode.FileLargest1000:
					return file.Rank != -1;
			}
			return false;
		}
	}
}
